! ------------------------------------------------------------------------
!>  \brief Module for Boundary Data
! ------------------------------------------------------------------------
MODULE boundary_data
USE system_constants      !> Global system constants
USE log_message_control   !> Printing log/error messages
USE log_messages          !> Log/error message and sender codes
USE field_data            !> Field Data module

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported data types
PUBLIC :: surfLoadT


!> Exported interfaces
PUBLIC :: bnd_init, bnd_clean, &
          bnd_numBoundElem, &
          bnd_getConnect, bnd_setConnect, &
          bnd_lenBoundElem, &
          bnd_getTrac, bnd_setTrac



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

!> sender code for this module
INTEGER, PARAMETER :: sdr = BNDDAT



! ************************************************************************
! DATA TYPES
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Surface load
!!
!!  \param sig_nt   Shear stress
!!  \param sig_nn   Normal stress
! ------------------------------------------------------------------------
TYPE surfLoadT
  DOUBLE PRECISION :: sig_nt, sig_nn
END TYPE surfLoadT


! ------------------------------------------------------------------------
!>  \brief Surface traction
!!
!!  \param loads   Set of node loads
! ------------------------------------------------------------------------
TYPE tracT
  TYPE(surfLoadT), DIMENSION(NNODELB) :: loads
END TYPE tracT


! ------------------------------------------------------------------------
!>  \brief Connectivity
!!
!!  \param nd   Sequence of node indices
! ------------------------------------------------------------------------
TYPE boundConnectT
  INTEGER, DIMENSION(NNODELB) :: nd
END TYPE boundConnectT


! ------------------------------------------------------------------------
!>  \brief Traction Element
!!
!!  \param num      Element number
!!  \param ico      Connectivity of local nodes
!!  \param trac     Surface traction of element
! ------------------------------------------------------------------------
TYPE boundElementT
  INTEGER             :: num
  TYPE(boundConnectT) :: ico
  TYPE(tracT)         :: trac
END TYPE boundElementT



! ************************************************************************
! STATE VARIABLES
! ************************************************************************

TYPE(boundElementT), ALLOCATABLE :: boundElements(:)  !> set of traction elements



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to initializer for boundary element data
INTERFACE bnd_init
  MODULE PROCEDURE bnd_init_
  MODULE PROCEDURE bnd_init_exc_
END INTERFACE bnd_init

!> \brief Interface to destructor for boundary element data
INTERFACE bnd_clean
  MODULE PROCEDURE bnd_clean_
END INTERFACE bnd_clean

!> \brief Interface to getter for number of traction elements
INTERFACE bnd_numBoundElem
  MODULE PROCEDURE bnd_num_bound_elem_
END INTERFACE bnd_numBoundElem

!> \brief Interface to getter for traction element connectivity
INTERFACE bnd_getConnect
  MODULE PROCEDURE bnd_get_connect_
  MODULE PROCEDURE bnd_get_connect_exc_
END INTERFACE bnd_getConnect

!> \brief Interface to setter for traction element connectivity
INTERFACE bnd_setConnect
  MODULE PROCEDURE bnd_set_connect_
  MODULE PROCEDURE bnd_set_connect_exc_
END INTERFACE bnd_setConnect

!> \brief Interface to getter for length of traction element
INTERFACE bnd_lenBoundElem
  MODULE PROCEDURE bnd_len_bound_elem_
  MODULE PROCEDURE bnd_len_bound_elem_exc_
END INTERFACE bnd_lenBoundElem

!> \brief Interface to getter for tractions
INTERFACE bnd_getTrac
  MODULE PROCEDURE bnd_get_trac_
  MODULE PROCEDURE bnd_get_trac_exc_
END INTERFACE bnd_getTrac

!> \brief Interface to setter for tractions
INTERFACE bnd_setTrac
  MODULE PROCEDURE bnd_set_trac_
  MODULE PROCEDURE bnd_set_trac_exc_
END INTERFACE bnd_setTrac



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Constructor for traction element data (non-exception checking)
!!
!!  \param  nel  Number of elements
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of elements for the problem.
! ------------------------------------------------------------------------
SUBROUTINE bnd_init_ (nel)
  INTEGER, INTENT(IN) :: nel
  INTEGER :: i, j     !> loop variables

  !> ensure state variable is clear
  CALL bnd_clean()

  !> allocate memory for element data state variable
  ALLOCATE(boundElements(nel))

  !> initialize state variable
  DO i = 1,nel

    !> element number
    boundElements(i)%num   = i

    !> initialize connectivity and loads
    DO j = 1,NNODELB
      boundElements(i)%ico  %nd(j)            = 0
      boundElements(i)%trac %loads(j) %sig_nt = 0.d0
      boundElements(i)%trac %loads(j) %sig_nn = 0.d0
    END DO

  END DO

END SUBROUTINE bnd_init_


! ------------------------------------------------------------------------
!>  \brief Constructor for traction element data (exception checking)
!!
!!  \param  nel   Number of elements
!!  \param  exc   Error code
!!
!!  \exception  ALLOC Failed to allocate memory for element data
!!  \exception  SZE   Specified number of elements is invalid
!!
!!  This routine allocates memory and initializes the state variable that
!!  contains the set of elements for the problem.
! ------------------------------------------------------------------------
SUBROUTINE bnd_init_exc_ (nel, exc)
  INTEGER, INTENT(IN) :: nel
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: e        !> status code for memory allocation
  INTEGER :: i, j     !> loop variables

  !> ensure state variable is clear
  CALL bnd_clean()

  !> make sure number of elements is valid
  IF (nel.LT.1 .OR. nel.GT.MAX_BOUNDELS) THEN
    exc=SZE
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> allocate memory for element data state variable
  ALLOCATE(boundElements(nel), STAT=e)

  !> make sure memory allocation was successful
  IF (e.NE.0) THEN
    exc=ALLOC
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> initialize state variable
  DO i = 1,nel

    !> element number
    boundElements(i)%num   = i

    !> initialize connectivity and loads
    DO j = 1,NNODELB
      boundElements(i)%ico  %nd(j)            = 0
      boundElements(i)%trac %loads(j) %sig_nt = 0.d0
      boundElements(i)%trac %loads(j) %sig_nn = 0.d0
    END DO

  END DO

END SUBROUTINE bnd_init_exc_


! ------------------------------------------------------------------------
!>  \brief Destructor for traction element data
!!
!!  This routine clears the memory allocated to the state variable that
!!  contains the set of element data for the problem.
! ------------------------------------------------------------------------
SUBROUTINE bnd_clean_ ()

  IF (ALLOCATED(boundElements)) DEALLOCATE(boundElements)

END SUBROUTINE bnd_clean_


! ------------------------------------------------------------------------
!>  \brief Getter for number of traction elements
!!
!!  \return nel  Number of elements
!!
!!  This routine determines the number of data entries that have been
!!  allocated for element data. It does not check that the element data
!!  has been populated (i.e. changed from initial zero values).
! ------------------------------------------------------------------------
FUNCTION bnd_num_bound_elem_ () RESULT(nel)
  INTEGER :: nel

  !> if data is initialized, return number of elements
  IF (ALLOCATED(boundElements)) THEN
    nel = SIZE(boundElements)
  ELSE
    nel = 0   !> if not initialized, there are no traction elements
  END IF

END FUNCTION bnd_num_bound_elem_


! ------------------------------------------------------------------------
!>  \brief Getter for connectivity (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Local node index
!!
!!  \return n     Global node number
!!
!!  This routine determines the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
FUNCTION bnd_get_connect_ (i,j) RESULT(n)
  INTEGER, INTENT(IN) :: i,j
  INTEGER :: n

  n = boundElements(i)%ico%nd(j)

END FUNCTION bnd_get_connect_


! ------------------------------------------------------------------------
!>  \brief Getter for connectivity (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  exc   Error code
!!
!!  \return n     Global node number
!!
!!  \exception  POSIT The element number is not in [1..bnd_numBoundElem()]
!!                    or the local node index is not in [1..NNODELB]
!!
!!  This routine determines the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
FUNCTION bnd_get_connect_exc_ (i,j, exc) RESULT(n)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  INTEGER :: n

  !> check that the element number is within the bounds of the elements
  !! list and that the local node index is within [1..NNODELB]
  IF ( i.LT.1 .OR. i.GT.bnd_numBoundElem() &
        .OR. j.LT.1 .OR. j.GT.NNODELB ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    n = 0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  n = bnd_getConnect(i,j)

END FUNCTION bnd_get_connect_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for connectivity (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  n     Global node number
!!
!!  This routine sets the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
SUBROUTINE bnd_set_connect_ (i,j, n)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(IN) :: n

  boundElements(i)%ico%nd(j) = n

END SUBROUTINE bnd_set_connect_


! ------------------------------------------------------------------------
!>  \brief Setter for connectivity (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  n     Global node number
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The value of the node number exceeds the number of
!!                      nodes
!!  \exception  POSIT   The element number is not in
!!                      [1..bnd_numBoundElem()] or the local node index is
!!                      not in [1..NNODELB]
!!
!!  This routine sets the global node number associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
SUBROUTINE bnd_set_connect_exc_ (i,j, n, exc)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(IN) :: n
  INTEGER, INTENT(OUT) :: exc

  !> check that the element number is within the bounds of the elements
  !! list and that the local node index is within [1..NNODELB]
  IF ( i.LT.1 .OR. i.GT.bnd_numBoundElem() &
        .OR. j.LT.1 .OR. j.GT.NNODELB ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF (n.LT.1 .OR. n.GT.fld_numNode()) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL bnd_setConnect(i,j, n)

END SUBROUTINE bnd_set_connect_exc_


! ------------------------------------------------------------------------
!>  \brief Compute the length of an element (non-exception checking)
!!
!!  \param  i       Element number
!!
!!  \return length  Volume of element
!!
!!  This routine computes the length of a traction element. This is
!!  computed as:
!!
!!    L   =  ( (y2-y1)**2 + (x2-x1)**2 )**0.5
!!
!!  where xj and yj are the (x,y) coordinates of node j of element i.
! ------------------------------------------------------------------------
FUNCTION bnd_len_bound_elem_ (i) RESULT(length)
  INTEGER, INTENT(IN) :: i
  DOUBLE PRECISION :: length
  DOUBLE PRECISION, DIMENSION(NNODELB,NDIM) :: x  !> element coords
  INTEGER :: j,k    !> loop variable

  !> get element coords
  DO k = 1,NDIM
    DO j = 1,NNODELB
      x(j,k) = fld_getCoord(boundElements(i)%ico%nd(j), k)
    END DO
  END DO

  !> compute length
  length = SQRT( (x(2,2)-x(1,2))**2 + (x(2,1)-x(1,1))**2 )

END FUNCTION bnd_len_bound_elem_


! ------------------------------------------------------------------------
!>  \brief Compute the length of an element (exception checking)
!!
!!  \param  i     Element number
!!  \param  exc   Error code
!!
!!  \return length   Length of element
!!
!!  \exception  POSIT   The specified element number is not within the
!!                      range [1..bnd_numBoundElem()]
!!
!!  This routine computes the length of a traction element. This is
!!  computed as:
!!
!!    L   =  ( (y2-y1)**2 + (x2-x1)**2 )**0.5
!!
!!  where xj and yj are the (x,y) coordinates of node j of element i.
! ------------------------------------------------------------------------
FUNCTION bnd_len_bound_elem_exc_ (i, exc) RESULT(length)
  INTEGER, INTENT(IN) :: i
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: length

  !> make sure the element number is valid
  IF (i.LT.1 .OR. i.GT.bnd_numBoundElem()) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    length = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  length = bnd_lenBoundElem(i)

END FUNCTION bnd_len_bound_elem_exc_


! ------------------------------------------------------------------------
!>  \brief Getter for traction (non-exception checking)
!!
!!  \param  i     Element number
!!  \param  j     Local node index
!!
!!  \return load  Surface load
!!
!!  This routine determines the surface load associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
FUNCTION bnd_get_trac_ (i,j) RESULT(load)
  INTEGER, INTENT(IN) :: i,j
  TYPE(surfLoadT) :: load

  load%sig_nt = boundElements(i)%trac%loads(j)%sig_nt
  load%sig_nn = boundElements(i)%trac%loads(j)%sig_nn

END FUNCTION bnd_get_trac_


! ------------------------------------------------------------------------
!>  \brief Getter for traction (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  exc   Error code
!!
!!  \return load  Surface load
!!
!!  \exception  POSIT The element number is not in [1..bnd_numBoundElem()]
!!                    or the local node index is not in [1..NNODELB]
!!
!!  This routine determines the surface load associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
FUNCTION bnd_get_trac_exc_ (i,j, exc) RESULT(load)
  INTEGER, INTENT(IN) :: i,j
  INTEGER, INTENT(OUT) :: exc
  TYPE(surfLoadT) :: load

  !> check that the element number is within the bounds of the elements
  !! list and that the local node index is within [1..NNODELB]
  IF ( i.LT.1 .OR. i.GT.bnd_numBoundElem() &
        .OR. j.LT.1 .OR. j.GT.NNODELB ) THEN
    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    load%sig_nt = 0.d0
    load%sig_nn = 0.d0
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  load = bnd_getTrac(i,j)

END FUNCTION bnd_get_trac_exc_


! ------------------------------------------------------------------------
!>  \brief Setter for traction (non-exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  load  Surface load
!!
!!  This routine sets the surface load associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
SUBROUTINE bnd_set_trac_ (i,j, load)
  INTEGER, INTENT(IN) :: i,j
  TYPE(surfLoadT), INTENT(IN) :: load

  boundElements(i)%trac%loads(j)%sig_nt = load%sig_nt
  boundElements(i)%trac%loads(j)%sig_nn = load%sig_nn

END SUBROUTINE bnd_set_trac_


! ------------------------------------------------------------------------
!>  \brief Setter for traction (exception checking)
!!
!!  \param  i     Node number
!!  \param  j     Local node index
!!  \param  load  Surface load
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  A value in the set of surface load stress is not
!!                      within the specified limits
!!  \exception  POSIT   The element number is not in
!!                      [1..bnd_numBoundElem()] or the local node index is
!!                      not in [1..NNODELB]
!!
!!  This routine sets the surface load associated with local
!!  node j for element i.
! ------------------------------------------------------------------------
SUBROUTINE bnd_set_trac_exc_ (i,j, load, exc)
  INTEGER, INTENT(IN) :: i,j
  TYPE(surfLoadT), INTENT(IN) :: load
  INTEGER, INTENT(OUT) :: exc

  !> check that the element number is within the bounds of the elements
  !! list and that the local node index is within [1..NNODELB]
  IF ( i.LT.1 .OR. i.GT.bnd_numBoundElem() &
        .OR. j.LT.1 .OR. j.GT.NNODELB ) THEN

    exc=POSIT
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE IF ( load%sig_nt.LT.SIG_MIN .OR. load%sig_nt.GT.SIG_MAX &
            .OR. load%sig_nn.LT.SIG_MIN .OR. load%sig_nn.GT.SIG_MAX ) THEN

    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN

  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL bnd_setTrac(i,j, load)

END SUBROUTINE bnd_set_trac_exc_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE boundary_data
