! ------------------------------------------------------------------------
!>  \brief Module for Constitutive Matrix
! ------------------------------------------------------------------------
MODULE constitutive
USE system_constants    !> Global constants (for size of matrix)
USE log_message_control !> Printing log/error messages
USE log_messages        !> Log/error codes
USE dense_matrix_def    !> Dense Matrix data type

IMPLICIT NONE
PRIVATE



! ************************************************************************
! EXPORTS
! ************************************************************************


!> Exported interfaces
PUBLIC :: dmatrix



! ************************************************************************
! LOCAL CONSTANTS
! ************************************************************************

INTEGER, PARAMETER :: sdr=CNSMAT


! ************************************************************************
! DATA TYPES
! ************************************************************************

! none



! ************************************************************************
! INTERFACES
! ************************************************************************


!> \brief Interface to constitutive matrix access program
INTERFACE dmatrix
  MODULE PROCEDURE dmatrix_linear_elastic_plane_strain_
  MODULE PROCEDURE dmatrix_linear_elastic_plane_strain_exc_
END INTERFACE dmatrix



CONTAINS



! ************************************************************************
! ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Constitutive matrix (linear elastic, plane strain)
!!
!!  \param  emod  Elastic modulus
!!  \param  nu    Poisson's ratio
!!  \param  Dmat  Constitutive matrix
!!
!!  This routine builds the constitutive matrix for a linear elastic
!!  material under plane strain conditions. The matrix has the following
!!  form:
!!
!!    D = ( emod / ((1+nu)*(1-2*nu)) ) * [ 1-nu   nu        0       ]
!!                                       [  nu   1-nu       0       ]
!!                                       [   0     0   0.5*(1-2*v)  ]
! ------------------------------------------------------------------------
SUBROUTINE dmatrix_linear_elastic_plane_strain_ (emod,nu, Dmat)
  DOUBLE PRECISION, INTENT(IN) :: emod,nu
  TYPE(matrixT), INTENT(INOUT) :: Dmat
  DOUBLE PRECISION :: one_minus_nu, one_minus_two_nu, coef

  one_minus_nu = 1.d0 - nu
  one_minus_two_nu = 1.d0 - 2.d0*nu
  coef = emod / ((1.d0+nu)*one_minus_two_nu)

  CALL dm_init(Dmat, NTNS,NTNS)

  CALL dm_set(Dmat, 1,1, coef*one_minus_nu)
  CALL dm_set(Dmat, 2,1, coef*nu)
  CALL dm_set(Dmat, 1,2, dm_get(Dmat, 2,1))
  CALL dm_set(Dmat, 2,2, dm_get(Dmat, 1,1))
  CALL dm_set(Dmat, 3,3, 0.5d0*coef*one_minus_two_nu)

END SUBROUTINE dmatrix_linear_elastic_plane_strain_


! ------------------------------------------------------------------------
!>  \brief Constitutive matrix (linear elastic, plane strain)
!!
!!  \param  emod  Elastic modulus
!!  \param  nu    Poisson's ratio
!!  \param  Dmat  Constitutive matrix
!!  \param  exc   Error code
!!
!!  \exception  EXCEED  The specified values of elastic modulus or
!!                      Poisson's ratio are not within specified limits
!!
!!  This routine builds the constitutive matrix for a linear elastic
!!  material under plane strain conditions. The matrix has the following
!!  form:
!!
!!    D = ( emod / ((1+nu)*(1-2*nu)) ) * [ 1-nu   nu        0       ]
!!                                       [  nu   1-nu       0       ]
!!                                       [   0     0   0.5*(1-2*v)  ]
! ------------------------------------------------------------------------
SUBROUTINE dmatrix_linear_elastic_plane_strain_exc_ (emod,nu, Dmat, exc)
  DOUBLE PRECISION, INTENT(IN) :: emod,nu
  TYPE(matrixT), INTENT(INOUT) :: Dmat
  INTEGER, INTENT(OUT) :: exc
  DOUBLE PRECISION :: one_minus_nu, one_minus_two_nu, coef

  !> check that specified parameters are within limits
  IF (  emod.LT.E_MIN .OR. emod.GT.E_MAX  &
        .OR. nu.LT.NU_MIN .OR. nu.GT.NU_MAX ) THEN
    exc=EXCEED
    CALL log_printLogMsg(exc,sdr)
    RETURN
  ELSE
    exc=OK
  END IF

  !> call non-exception version
  CALL dmatrix(emod,nu, Dmat)

END SUBROUTINE dmatrix_linear_elastic_plane_strain_exc_



! ************************************************************************
! LOCAL FUNCTIONS
! ************************************************************************

! none

END MODULE constitutive
