! ------------------------------------------------------------------------
!>  \brief Module for testing Dense Matrix data type
! ------------------------------------------------------------------------
MODULE dense_matrix_test
USE fruit                   !> Unit testing framework
USE log_message_control     !> Printing log/error messages
USE log_messages            !> Log/error codes
USE dense_matrix_def        !> Dense Matrix data type
USE vector_def              !> Vector data type

IMPLICIT NONE


CONTAINS



! ------------------------------------------------------------------------
!>  \test Test for OK exception message on allocation
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param testName Filename for log file (required for exceptions)
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!
!!  This test initializes a matrixT and makes sure that the exception
!!  message is OK (i.e. allocation did not fail).
! ------------------------------------------------------------------------
SUBROUTINE test_dm_allocation_MSG
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_allocation_MSG'
  TYPE(matrixT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg=OK
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log message file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize matrix and check the exception
  CALL dm_init(test, m,n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate matrix
  CALL log_closeLogFile()
  CALL dm_clean(test)

END SUBROUTINE test_dm_allocation_MSG


! ------------------------------------------------------------------------
!>  \test Test for SZE exception message on allocation
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param testName Filename for log file (required for exceptions)
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!
!!  This test attempts to initialize a matrixT with invalid size
!!  parameters and verifies that the correct exception is returned.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_allocation_SZE
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_allocation_SZE'
  TYPE(matrixT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg=SZE
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> try to initialize with m=-1
  CALL dm_init(test, -1,n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to initialize with n=-1
  CALL dm_init(test, m,-1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate matrix
  CALL log_closeLogFile()
  CALL dm_clean(test)

END SUBROUTINE test_dm_allocation_SZE


! ------------------------------------------------------------------------
!>  \test Test for initialization of matrix data to zeros.
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expDat   Expected initial contents of matrix data
!!
!!  This test initializes a matrixT and ensures that the data is
!!  initialized to zeros.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_allocation_DAT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_allocation_DAT'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up the expected data
  expDat = RESHAPE(  (/ 0.d0,0.d0,0.d0,0.d0,0.d0,0.d0 /), SHAPE(expDat) )

  !> initialize matrix and test actual data
  CALL dm_init(test, m,n)
  CALL assertEquals(expDat,test%dat, m,n)

  !> deallocate matrix
  CALL dm_clean(test)

END SUBROUTINE test_dm_allocation_DAT


! ------------------------------------------------------------------------
!>  \test Test for deallocation
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test dense matrix
!!  \param m          Number of rows in the matrix
!!  \param n          Number of rows/columns in the matrix
!!  \param expBefore  Expected allocation status before deallocation
!!  \param expAfter   Expected allocation status after deallocation
!!  \param actBefore  Actual allocation status before deallocation
!!  \param actAfter   Actual allocation status after deallocation
!!
!!  This test initializes a matrixT and ensures that it is
!!  allocated. It then deallocates the object and ensures that it has
!!  been deallocated.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_deallocation
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_deallocation'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: m=3,n=2
  LOGICAL, PARAMETER :: expBefore = .TRUE.
  LOGICAL, PARAMETER :: expAfter = .FALSE.
  LOGICAL :: actBefore, actAfter

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the matrix
  CALL dm_init(test, m,n)

  !> make sure dat is allocated
  actBefore = ALLOCATED(test%dat)
  CALL assertEquals(expBefore,actBefore)

  !> deallocate the matrix
  CALL dm_clean(test)

  !> make sure dat is deallocated
  actAfter = ALLOCATED(test%dat)
  CALL assertEquals(expAfter,actAfter)

END SUBROUTINE test_dm_deallocation


! ------------------------------------------------------------------------
!>  \test Test for number of rows when matrix is not allocated
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test dense matrix
!!  \param expected   Expected number of rows
!!  \param actual     Actual number of rows
!!
!!  This test makes sure that the number of rows is returned as 0 when
!!  the matrix is not allocated.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_num_rows_not_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_num_rows_not_allocated'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: expected = 0
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> check number of rows
  actual = dm_numRows(test)
  CALL assertEquals(expected, actual)

END SUBROUTINE test_dm_num_rows_not_allocated


! ------------------------------------------------------------------------
!>  \test Test for number of rows when matrix is allocated
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test dense matrix
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expected   Expected number of rows
!!  \param actual     Actual number of rows
!!
!!  This test checks that the number of rows returned is correct when the
!!  matrix is allocated.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_num_rows_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_num_rows_allocated'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expected = m
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the matrix
  CALL dm_init(test, m,n)

  !> check number of rows
  actual = dm_numRows(test)
  CALL assertEquals(expected, actual)

  !> deallocate the matrix
  CALL dm_clean(test)

END SUBROUTINE test_dm_num_rows_allocated


! ------------------------------------------------------------------------
!>  \test Test for number of columns when matrix is not allocated
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test dense matrix
!!  \param expected   Expected number of columns
!!  \param actual     Actual number of columns
!!
!!  This test makes sure that the number of columns is returned as 0 when
!!  the matrix is not allocated
! ------------------------------------------------------------------------
SUBROUTINE test_dm_num_cols_not_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_num_cols_not_allocated'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: expected = 0
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> check number of rows
  actual = dm_numCols(test)
  CALL assertEquals(expected, actual)

END SUBROUTINE test_dm_num_cols_not_allocated


! ------------------------------------------------------------------------
!>  \test Test for number of columns when matrix is allocated
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test       Test dense matrix
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expected   Expected number of columns
!!  \param actual     Actual number of columns
!!
!!  This test checks that the number of columns returned is correct when
!!  the matrix is allocated.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_num_cols_allocated
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_num_cols_allocated'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expected = n
  INTEGER :: actual

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the matrix
  CALL dm_init(test, m,n)

  !> check number of columns
  actual = dm_numCols(test)
  CALL assertEquals(expected, actual)

  !> deallocate the matrix
  CALL dm_clean(test)

END SUBROUTINE test_dm_num_cols_allocated


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from dm_get
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param testName Filename for log file (required for exceptions)
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!  \param v        Dummy variable for get function return
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the matrix.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_get_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_get_POSIT'
  TYPE(matrixT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg = POSIT
  INTEGER :: actMsg
  DOUBLE PRECISION :: v

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize matrix
  CALL dm_init(test, m,n)

  !> try to get beyond last row
  v = dm_get(test, m+1,n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first row
  v = dm_get(test, 0,n, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get after last column
  v = dm_get(test, m,n+1, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to get before first column
  v = dm_get(test, m,0, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize the log file and deallocate the matrix
  CALL log_closeLogFile()
  CALL dm_clean(test)

END SUBROUTINE test_dm_get_POSIT


! ------------------------------------------------------------------------
!>  \test Test for correct value return from bsm_get
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param m        Number of rows in the matrix
!!  \param n        Number of rows/columns in the matrix
!!  \param i        Row index for test location
!!  \param j        Column index for test location
!!  \param expVal   Expected data value
!!  \param actVal   Actual data value
!!
!!  This test checks that the correct value is returned from the get
!!  access program.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_get_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_get_VAL'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: i=1,j=2
  DOUBLE PRECISION, PARAMETER :: expVal = 3.d0
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the matrix
  CALL dm_init(test, m,n)

  !> set the value manually in the internal data structure
  test%dat(i,j) = expVal

  !> get the value using the access program
  actVal = dm_get(test, i,j)
  CALL assertEquals(expVal,actVal)

  !> deallocate the matrix
  CALL dm_clean(test)

END SUBROUTINE test_dm_get_VAL


! ------------------------------------------------------------------------
!>  \test Test for POSIT exception from dm_set
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param testName Filename for log file (required for exceptions)
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expMsg   Expected log message
!!  \param actMsg   Actual log message
!!  \param v        Dummy variable for set function input
!!
!!  This test checks that a POSIT exception is returned when the requested
!!  location is not inside the matrix.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_set_POSIT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_set_POSIT'
  TYPE(matrixT) :: test
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg = POSIT
  DOUBLE PRECISION, PARAMETER :: v=0.d0
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize the matrix
  CALL dm_init(test, m,n)

  !> try to set after last row
  CALL dm_set(test, m+1,n,v, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first row
  CALL dm_set(test, 0,n,v, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set after last column
  CALL dm_set(test, m,n+1,v, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> try to set before first column
  CALL dm_set(test, m,0,v, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate matrix
  CALL log_closeLogFile()
  CALL dm_clean(test)

END SUBROUTINE test_dm_set_POSIT


! ------------------------------------------------------------------------
!>  \test Test for correct value return after dm_set
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test     Test dense matrix
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param i        Row index for test location
!!  \param j        Column index for test location
!!  \param expVal   Expected data value
!!  \param actVal   Actual data value
!!
!!  This test checks that the correct value is entered by the set
!!  access program.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_set_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_set_VAL'
  TYPE(matrixT) :: test
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: i=1,j=2
  DOUBLE PRECISION, PARAMETER :: expVal = 3.d0
  DOUBLE PRECISION :: actVal

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize the matrix
  CALL dm_init(test, m,n)

  !> set the value
  CALL dm_set(test, i,j,expVal)
  actVal = dm_get(test, i,j)
  CALL assertEquals(expVal,actVal)

  !> deallocate matrix
  CALL dm_clean(test)

END SUBROUTINE test_dm_set_VAL


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception when adding matrices
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First input dense matrix
!!  \param test2      Second input dense matrix
!!  \param test3      Output dense matrix
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!
!!  This test checks that a DIMEN exception is raised when matrices with
!!  incompatible dimensions are added.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_add_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_add_DIMEN'
  TYPE(matrixT) :: test1, test2, test3
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg=DIMEN
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize input matrices (data contents not important)
  CALL dm_init(test1, m,n)
  CALL dm_init(test2, m+1,n)  !> note different number of rows

  !> attempt to add the matrices
  test3 = dm_add(test1,test2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> re-initialize input matrices (data contents not important)
  CALL dm_init(test1, m,n)
  CALL dm_init(test2, m,n+1)  !> note different number of columns

  !> attempt to add the matrices
  test3 = dm_add(test1,test2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate matrices
  CALL log_closeLogFile()
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_add_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for correct addition of matrices
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First input dense matrix
!!  \param test2      Second input dense matrix
!!  \param test3      Output dense matrix
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expDat     Expected result of addition
!!
!!  This test checks that the correct result is obtained when adding two
!!  dense matrices.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_add_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_add_VAL'
  TYPE(matrixT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 3.d0,3.d0,3.d0,3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize first input matrix
  !! (loop sets matrix to all ones)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> initialize second input matrix
  !! (loop sets matrix to all twos)
  CALL dm_init(test2, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test2, i,j, 2.d0)
    END DO
  END DO

  !> add the matrices
  test3 = dm_add(test1,test2)
  CALL assertEquals(expDat,test3%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_add_VAL


! ------------------------------------------------------------------------
!>  \test Test for correct addition of matrices ( with OPERATOR (+) )
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First input dense matrix
!!  \param test2      Second input dense matrix
!!  \param test3      Output dense matrix
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expDat     Expected result of addition
!!
!!  This test checks that the correct result is obtained when adding two
!!  dense matrices using the (+) operator.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_add_OP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_add_OP'
  TYPE(matrixT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 3.d0,3.d0,3.d0,3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize first input matrix
  !! (loop sets matrix to all ones)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> initialize second input matrix
  !! (loop sets matrix to all twos)
  CALL dm_init(test2, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test2, i,j, 2.d0)
    END DO
  END DO

  !> add the matrices
  test3 = test1 + test2
  CALL assertEquals(expDat,test3%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_add_OP


! ------------------------------------------------------------------------
!>  \test Test for scalar multiplication
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    Input dense matrix
!!  \param test2    Output dense matrix
!!  \param k        Scalar coefficient
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expDat   Expected data contents
!!
!!  This test checks that scalar multiplication is performed correctly.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_scal_mul_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_scal_mul_VAL'
  TYPE(matrixT) :: test1, test2
  DOUBLE PRECISION, PARAMETER :: k=3.d0
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 3.d0,3.d0,3.d0,3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize matrix
  !! (loop sets matrix to all ones)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> multiply by scalar
  test2 = dm_scalMul(test1,k)
  CALL assertEquals(expDat,test2%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)

END SUBROUTINE test_dm_scal_mul_VAL


! ------------------------------------------------------------------------
!>  \test Test for scalar multiplication (by zero)
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    Input dense matrix
!!  \param test2    Output dense matrix
!!  \param k        Scalar coefficient (zero)
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expDat   Expected data contents
!!
!!  This test checks that scalar multiplication is performed correctly.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_scal_mul_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_scal_mul_ZERO'
  TYPE(matrixT) :: test1, test2
  DOUBLE PRECISION, PARAMETER :: k=0.d0
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 0.d0,0.d0,0.d0,0.d0,0.d0,0.d0 /), SHAPE(expDat) )

  !> initialize matrix
  !! (loop sets matrix to all ones)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> multiply by zero
  test2 = dm_scalMul(test1,k)
  CALL assertEquals(expDat,test2%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)

END SUBROUTINE test_dm_scal_mul_ZERO


! ------------------------------------------------------------------------
!>  \test Test for scalar multiplication ( using OPERATOR (*) )
!!
!!  \param unit_name  Name of unit test (for FRUIT)
!!  \param test1    Input dense matrix
!!  \param test2    Output dense matrix
!!  \param k        Scalar coefficient (zero)
!!  \param m        Number of rows in the matrix
!!  \param n        Number of columns in the matrix
!!  \param expDat   Expected data contents
!!
!!  This test checks that scalar multiplication is performed correctly
!!  when using the overloaded (*) operator.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_scal_mul_OP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_scal_mul_OP'
  TYPE(matrixT) :: test1, test2
  DOUBLE PRECISION, PARAMETER :: k=3.d0
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize expected result
  expDat = RESHAPE(  (/ 3.d0,3.d0,3.d0,3.d0,3.d0,3.d0 /), SHAPE(expDat) )

  !> initialize matrix
  !! (loop sets matrix to all ones)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> multiply by scalar
  test2 = k*test1
  CALL assertEquals(expDat,test2%dat, m,n)

  !> clear output matrix
  CALL dm_clean(test2)

  !> multiply by scalar (switch order of inputs)
  test2 = test1*k
  CALL assertEquals(expDat,test2%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)

END SUBROUTINE test_dm_scal_mul_OP


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception when performing matrix-vector multiplication
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Input dense matrix
!!  \param test2      Input vector
!!  \param test3      Dummy output vector
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!
!!  This test checks that a DIMEN exception is raised when matrix-vector
!!  multiplication is attempted with a vector whose length does not equal
!!  the number of rows in the matrix.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_vec_mul_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_vec_mul_DIMEN'
  TYPE(matrixT) :: test1
  TYPE(vectorT) :: test2, test3
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg=DIMEN
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> set up matrix and vector (data contents not important)
  CALL dm_init(test1, m,n)
  CALL vec_init(test2, m)

  !> attempt matrix-vector multiplication
  test3 = dm_vecMul(test1,test2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate objects
  CALL log_closeLogFile()
  CALL dm_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_dm_vec_mul_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for correct matrix-vector multiplication
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Input dense matrix
!!  \param test2      Input vector
!!  \param test3      Output vector
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-vector multiplication is performed
!!  correctly in the general case (input matrix and vector contain data,
!!  but it is not of a special form).
! ------------------------------------------------------------------------
SUBROUTINE test_dm_vec_mul_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_vec_mul_VAL'
  TYPE(matrixT) :: test1
  TYPE(vectorT) :: test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set expected result
  expDat = RESHAPE(  (/ 9.d0,12.d0,15.d0 /), SHAPE(expDat) )

  !> initialize input matrix
  !! (loop sets up data)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> initialize input vector
  !! (loop sets up data)
  CALL vec_init(test2, n)
  DO i = 1,n
    CALL vec_set(test2, i, DBLE(i))
  END DO

  !> perform multiplication
  test3 = dm_vecMul(test1,test2)
  CALL assertEquals(expDat,test3%dat, m)

  !> deallocate objects
  CALL dm_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_dm_vec_mul_VAL


! ------------------------------------------------------------------------
!>  \test Test for matrix-vector multiplication (by the zero vector)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Input dense matrix
!!  \param test2      Input vector (zeros)
!!  \param test3      Output vector
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-vector multiplication is performed
!!  correctly when the input vector is all zeros.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_vec_mul_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_vec_mul_ZERO'
  TYPE(matrixT) :: test1
  TYPE(vectorT) :: test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set expected result
  expDat = RESHAPE(  (/ 0.d0,0.d0,0.d0 /), SHAPE(expDat) )

  !> initialize input matrix
  !! (loop sets up data)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> initialize input vector (all zeros)
  CALL vec_init(test2, n)

  !> perform multiplication
  test3 = dm_vecMul(test1,test2)
  CALL assertEquals(expDat,test3%dat, m)

  !> deallocate objects
  CALL dm_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_dm_vec_mul_ZERO


! ------------------------------------------------------------------------
!>  \test Test for matrix-vector multiplication (by the identity matrix)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Input dense matrix (identity)
!!  \param test2      Input vector
!!  \param test3      Output vector
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-vector multiplication is performed
!!  correctly when the input matrix is the identity for multiplication.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_vec_mul_IDENT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_vec_mul_IDENT'
  TYPE(matrixT) :: test1
  TYPE(vectorT) :: test2, test3
  INTEGER, PARAMETER :: m=3
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize input matrix
  !! (loop sets up identity matrix)
  CALL dm_init(test1, m,m)
  DO j = 1,m
    DO i = 1,m
      IF (i.EQ.j) CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> initialize input vector
  !! (loop sets data)
  CALL vec_init(test2, m)
  DO i = 1,m
    CALL vec_set(test2, i, DBLE(i))
  END DO

  !> expected result is the same as the initial vector
  expDat = test2%dat

  !> perform multiplication
  test3 = dm_vecMul(test1,test2)
  CALL assertEquals(expDat,test3%dat, m)

  !> deallocate objects
  CALL dm_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_dm_vec_mul_IDENT


! ------------------------------------------------------------------------
!>  \test Test for correct matrix-vector multiplication ( using OPERATOR (*) )
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Input dense matrix
!!  \param test2      Input vector
!!  \param test3      Output vector
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-vector multiplication is performed
!!  correctly in the general case (input matrix and vector contain data,
!!  but it is not of a special form). The overloaded operator (*) is used.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_vec_mul_OP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_vec_mul_OP'
  TYPE(matrixT) :: test1
  TYPE(vectorT) :: test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected result
  expDat = RESHAPE(  (/ 9.d0,12.d0,15.d0 /), SHAPE(expDat) )

  !> initialize input matrix
  !! (loop sets data)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> initialize input vector
  !! (loop sets data)
  CALL vec_init(test2, n)
  DO i = 1,n
    CALL vec_set(test2, i, DBLE(i))
  END DO

  !> perform multiplication
  test3 = test1*test2
  CALL assertEquals(expDat,test3%dat, m)

  !> deallocate objects
  CALL dm_clean(test1)
  CALL vec_clean(test2)
  CALL vec_clean(test3)

END SUBROUTINE test_dm_vec_mul_OP


! ------------------------------------------------------------------------
!>  \test Test for DIMEN exception when performing matrix-matrix multiplication
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First (left-hand) input dense matrix
!!  \param test2      Second (right-hand) input dense matrix
!!  \param test3      Dummy output matric
!!  \param testName   Filename for log file (required for exceptions)
!!  \param m          Number of rows in the matrix
!!  \param n          Number of columns in the matrix
!!  \param expMsg     Expected exception message
!!  \param actMsg     Actual exception message
!!
!!  This test checks that a DIMEN exception is raised when matrix-matrix
!!  multiplication is attempted between matrices whose dimensions are not
!!  compatible (i.e. number of columns in test1 does not equal number of
!!  rows in test2).
! ------------------------------------------------------------------------
SUBROUTINE test_dm_mat_mul_DIMEN
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_mat_mul_DIMEN'
  TYPE(matrixT) :: test1, test2, test3
  CHARACTER (LEN=*), PARAMETER :: testName = 'testName'
  INTEGER, PARAMETER :: m=3,n=2
  INTEGER, PARAMETER :: expMsg=DIMEN
  INTEGER :: actMsg

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize log file
  CALL log_setFileName(testName)
  CALL log_initLogFile()

  !> initialize matrices (data content not important)
  CALL dm_init(test1, m,n)
  CALL dm_init(test2, m,n)

  !> attempt matrix-matrix multiplication
  test3 = dm_matMul(test1,test2, actMsg)
  CALL assertEquals(expMsg,actMsg)

  !> finalize log file and deallocate matrices
  CALL log_closeLogFile()
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_mat_mul_DIMEN


! ------------------------------------------------------------------------
!>  \test Test for correct matrix-matrix multiplication
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First (left-hand) input dense matrix
!!  \param test2      Second (right-hand) input dense matrix
!!  \param test3      Output matrix
!!  \param m          Number of rows in matrices
!!  \param n          Number of columns in matrices
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-matrix multiplication is performed
!!  correctly in the general case (input matrices contain data,
!!  but it is not of a special form).
! ------------------------------------------------------------------------
SUBROUTINE test_dm_mat_mul_VAL
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_mat_mul_VAL'
  TYPE(matrixT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected result
  expDat = RESHAPE( (/ &
                    30.d0,36.d0,42.d0,66.d0,81.d0,96.d0 &
                    /), SHAPE(expDat) )

  !> initialize first matrix (dimensions of m x m)
  !! (loops sets up data)
  CALL dm_init(test1, m,m)
  DO j = 1,m
    DO i = 1,m
      CALL dm_set(test1, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> initialize second matrix (dimensions of m x n)
  !! (loops sets up data)
  CALL dm_init(test2, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test2, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> perform matrix-matrix multiplication
  test3 = dm_matMul(test1,test2)
  CALL assertEquals(expDat,test3%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_mat_mul_VAL


! ------------------------------------------------------------------------
!>  \test Test for correct matrix-matrix multiplication (by zero matrix)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First (left-hand) input dense matrix (zeros)
!!  \param test2      Second (right-hand) input dense matrix
!!  \param test3      Output matrix
!!  \param m          Number of rows in matrices
!!  \param n          Number of columns in matrices
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-matrix multiplication is performed
!!  correctly for multiplication by a zero matrix.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_mat_mul_ZERO
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_mat_mul_ZERO'
  TYPE(matrixT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected result
  expDat = RESHAPE(  (/ 0.d0,0.d0,0.d0,0.d0,0.d0,0.d0 /), SHAPE(expDat) )

  !> initialize first matrix (zeros)
  CALL dm_init(test1, m,m)

  !> initialize second matrix
  !! (loop sets up data)
  CALL dm_init(test2, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test2, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> perform matrix-matrix multiplication
  test3 = dm_matMul(test1,test2)
  CALL assertEquals(expDat,test3%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_mat_mul_ZERO


! ------------------------------------------------------------------------
!>  \test Test for correct matrix-matrix multiplication (by identity matrix)
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First (left-hand) input dense matrix (identity)
!!  \param test2      Second (right-hand) input dense matrix
!!  \param test3      Output matrix
!!  \param m          Number of rows in matrices
!!  \param n          Number of columns in matrices
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-matrix multiplication is performed
!!  correctly for multiplication by the identity matrix for multiplication.
! ------------------------------------------------------------------------
SUBROUTINE test_dm_mat_mul_IDENT
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_mat_mul_IDENT'
  TYPE(matrixT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize first matrix
  !! (loop sets up identity matrix)
  CALL dm_init(test1, m,m)
  DO j = 1,m
    DO i = 1,m
      IF (i.EQ.j) CALL dm_set(test1, i,j, 1.d0)
    END DO
  END DO

  !> initialize second matrix
  !! (loop sets up data)
  CALL dm_init(test2, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test2, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> expected result is the same as the second matrix
  expDat = test2%dat

  !> perform matrix-matrix multiplication
  test3 = dm_matMul(test1,test2)
  CALL assertEquals(expDat,test3%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_mat_mul_IDENT


! ------------------------------------------------------------------------
!>  \test Test for correct matrix-matrix multiplication ( using OPERATOR (*) )
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      First (left-hand) input dense matrix
!!  \param test2      Second (right-hand) input dense matrix
!!  \param test3      Output matrix
!!  \param m          Number of rows in matrices
!!  \param n          Number of columns in matrices
!!  \param expDat     Expected result
!!
!!  This test checks that matrix-matrix multiplication is performed
!!  correctly in the general case (input matrices contain data,
!!  but it is not of a special form). This test uses the overloaded
!!  operator (*).
! ------------------------------------------------------------------------
SUBROUTINE test_dm_mat_mul_OP
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_mat_mul_OP'
  TYPE(matrixT) :: test1, test2, test3
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(m,n) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> set up expected result
  expDat = RESHAPE( (/ &
                    30.d0,36.d0,42.d0,66.d0,81.d0,96.d0 &
                    /), SHAPE(expDat) )

  !> initialize first matrix (dimensions of m x m)
  !! (loop sets up data)
  CALL dm_init(test1, m,m)
  DO j = 1,m
    DO i = 1,m
      CALL dm_set(test1, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> initialize second matrix (dimensions of m x n)
  !! (loop sets up data)
  CALL dm_init(test2, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test2, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> perform matrix-matrix multiplication
  test3 = test1*test2
  CALL assertEquals(expDat,test3%dat, m,n)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)
  CALL dm_clean(test3)

END SUBROUTINE test_dm_mat_mul_OP


! ------------------------------------------------------------------------
!>  \test Test for matrix transpose operation
!!
!!  \param unit_name    Name of unit test (for FRUIT)
!!  \param test1      Input dense matrix
!!  \param test2      Output dense matrix
!!  \param m          Number of rows in matrices
!!  \param n          Number of columns in matrices
!!  \param expDat     Expected result
! ------------------------------------------------------------------------
SUBROUTINE test_dm_transpose
  CHARACTER (LEN=*), PARAMETER :: unit_name = 'test_dm_transpose'
  TYPE(matrixT) :: test1, test2
  INTEGER, PARAMETER :: m=3,n=2
  DOUBLE PRECISION, DIMENSION(n,m) :: expDat
  INTEGER :: i,j    !> loop variables

  !> initialize unit test
  CALL set_unit_name(unit_name)

  !> initialize input matrix
  !! (loop sets up data)
  CALL dm_init(test1, m,n)
  DO j = 1,n
    DO i = 1,m
      CALL dm_set(test1, i,j, ( (DBLE(j)-1.d0)*DBLE(m) + DBLE(i) ) )
    END DO
  END DO

  !> calculate expected result (using Fortran built-in)
  expDat = TRANSPOSE(test1%dat)

  !> perform transpose operation
  test2 = dm_transpose(test1)
  CALL assertEquals(expDat,test2%dat, n,m)

  !> deallocate matrices
  CALL dm_clean(test1)
  CALL dm_clean(test2)

END SUBROUTINE test_dm_transpose

END MODULE dense_matrix_test
