! ------------------------------------------------------------------------
!>  \brief Module defining Log Messages
! ------------------------------------------------------------------------
MODULE log_messages
USE system_constants    !> Global system constants (for max string length)

IMPLICIT NONE



! ************************************************************************
! EXPORTED DATA TYPES
! ************************************************************************


!> \brief Enumerated type messageT
ENUM, BIND(C)
  ENUMERATOR :: OK=1, ALLOC=2, DIMEN=3, EXCEED=4, EXISTS=5, FORMT=6
  ENUMERATOR :: POSIT=7, POSDEF=8, SZE=9, TYP=10
END ENUM ! messageT

!> \brief Enumerated type senderT
ENUM, BIND(C)
  ENUMERATOR :: BFCRDR=1, BNDDAT=2, BNDRDR=3, BSYMAT=4, CNSMAT=5, DMNRDR=6
  ENUMERATOR :: DNSMAT=7, FLDDAT=8, ICTRDR=9, ICVRDR=10, KBCRDR=11
  ENUMERATOR :: LINSLV=12, MTLDAT=13, MTLRDR=14, NBCRDR=15, TNSWTR=16
  ENUMERATOR :: VECTOR=17, VECWTR=18
END ENUM ! senderT



CONTAINS



! ************************************************************************
! EXPORTED ACCESS PROGRAMS
! ************************************************************************


! ------------------------------------------------------------------------
!>  \brief Function for getting the string message corresponding to a messageT code
!!
!!  \param  code Message code (messageT)
!!  \return exc  Human readable string representing exception
! ------------------------------------------------------------------------
FUNCTION msg_getMsg (code) RESULT(exc)
  INTEGER, INTENT(IN) :: code
  CHARACTER(LEN=MAXLEN) :: exc

  SELECT CASE (code)
    CASE (ALLOC)
      exc = 'Failed to allocate memory'
    CASE (DIMEN)
      exc = 'Size of data structure does not match size of target data structure'
    CASE (EXCEED)
      exc = 'Data value exceeds a defined minimum or maximum'
    CASE (EXISTS)
      exc = 'File does not exist'
    CASE (FORMT)
      exc = 'File data is not in expected format'
    CASE (POSDEF)
      exc = 'Matrix is not positive definite'
    CASE (POSIT)
      exc = 'Index exceeds size of data structure'
    CASE (SZE)
      exc = 'Specified size of data structure exceeds minimum or maximum allowable size'
    CASE (TYP)
      exc = 'Specified material property does not correspond to material type'
    CASE DEFAULT
      exc = 'Could not find log message code'
  END SELECT

END FUNCTION msg_getMsg


! ------------------------------------------------------------------------
!>  \brief Function for getting the string message corresponding to a senderT code
!!
!!  \param  code Message code (senderT)
!!  \return sdr  Human readable string representing sender
! ------------------------------------------------------------------------
FUNCTION msg_getSdr (code) RESULT(sdr)
  INTEGER, INTENT(IN) :: code
  CHARACTER(LEN=MAXLEN) :: sdr

  SELECT CASE (code)
    CASE (BFCRDR)
      sdr = 'Body Force Reader'
    CASE (BNDDAT)
      sdr = 'Boundary Data'
    CASE (BNDRDR)
      sdr = 'Boundary File Reader'
    CASE (BSYMAT)
      sdr = 'Banded Symmetric Matrix'
    CASE (CNSMAT)
      sdr = 'Constitutive Matrix'
    CASE (DMNRDR)
      sdr = 'Domain File Reader'
    CASE (DNSMAT)
      sdr = 'Dense Matrix'
    CASE (FLDDAT)
      sdr = 'Field Data'
    CASE (ICTRDR)
      sdr = 'Initial Tensor Field Reader'
    CASE (ICVRDR)
      sdr = 'Initial Vector Field Reader'
    CASE (KBCRDR)
      sdr = 'Kinematic BC Reader'
    CASE (LINSLV)
      sdr = 'Linear Solver'
    CASE (MTLDAT)
      sdr = 'Material Property Data'
    CASE (MTLRDR)
      sdr = 'Material File Reader'
    CASE (NBCRDR)
      sdr = 'Natural BC Reader'
    CASE (TNSWTR)
      sdr = 'Tensor Field Writer'
    CASE (VECTOR)
      sdr = 'Vector Data Type'
    CASE (VECWTR)
      sdr = 'Vector Field Writer'
    CASE DEFAULT
      sdr = 'Could not find sender code'
  END SELECT

END FUNCTION msg_getSdr

END MODULE log_messages
